|
|     Amiga Linux/68k Loader
|
|      Copyright 1995 by Geert Uytterhoeven
|			 (Geert.Uytterhoeven@cs.kuleuven.ac.be)
|
|     --------------------------------------------------------------------------
|
|     Cache fixes for V36 (A3000 Beta ROM) and some code optimizations
|     by Janos Farkas (chexum@sparta.banki.hu)
|
|     --------------------------------------------------------------------------
|
|     The original idea for this program comes from the Linux distribution:
|
|	 LILO - Generic Boot Loader for Linux ("LInux LOader")
|		by Werner Almesberger
|
|     --------------------------------------------------------------------------
|
|     This file is subject to the terms and conditions of the GNU General Public
|     License.  See the file README.legal in the main directory of the Linux/68k
|     distribution for more details.
|


.text
.globl _BootCodeStart, _BootCodeEnd, _LoaderVectorStart, _LoaderVectorEnd
.align 4


|	Alert Codes

AT_DeadEnd		= 0x80000000
AN_LILO			= AT_DeadEnd+0x4c000000
AG_NoMemory		= 0x00010000
AG_IOError		= 0x00060000
AG_BadParm		= 0x00080000
NOT_ENOUGH_MEMORY	= AN_LILO+AG_NoMemory	| Yields Alert Code CC010000
READ_ERROR		= AN_LILO+AG_IOError	| Yields Alert Code CC060000
HUNK_ERROR		= AN_LILO+AG_BadParm	| Yields Alert Code CC080000

|	Memory Allocation

MEMF_PUBLIC		= 1
MEMF_CLEAR		= 0x10000
MEMF_PUB_CLR		= MEMF_PUBLIC+MEMF_CLEAR

|	Exec I/O

IO_COMMAND		= 28
IO_FLAGS		= 30
IO_LENGTH		= 36
IO_DATA			= 40
IO_OFFSET		= 44
CMD_READ		= 2
IOF_QUICK		= 1

|	Trackdisk

TD_SECTOR		= 512
TD_SECSHIFT		= 9
TD_MOTOR		= 9

|	Library

LIB_VERSION		= 20

|	Resident Tag

RT_INIT			= 22

|	Expansion Library

eb_Flags		= 34
EBB_SILENTSTART		= 6

|	Exec Library

AttnFlags		= 0x128
AFB_68020		= 1
AFF_68040		= 8
AFF_FPU40		= 0x40

|	DOS Hunks

HUNK_CODE		= 1001
HUNK_DATA		= 1002
HUNK_BSS		= 1003
HUNK_RELOC32		= 1004
HUNK_END		= 1010
HUNK_HEADER		= 1011

|	Exec Library Vector Offsets

_LVOSuperVisor		= -0x1e
_LVOFindResident	= -0x60
_LVOAlert		= -0x6c
_LVOCloseLibrary	= -0x19e
_LVODoIO		= -0x1c8
_LVOOpenLibrary		= -0x228
_LVOCopyMem		= -0x270
_LVOCacheClearU		= -0x27c		| only in/after V37
_LVOAllocVec		= -0x2ac
_LVOFreeVec		= -0x2b2

|
|	Lilo data offsets (on the stack)
|

SysBase			=    0*4		| execbase
FuncTable		=    1*4		| Function table
BootRequest		=    7*4		| passed by bootstrap
FileVector		=    8*4
FileOffset		=    9*4
FileBuffer		=   10*4
BASE_SIZE		=   11*4

|
|	Boot Code
|
|		in:	a1 = bootrequest
|			a6 = execbase
|
|		out:	d0 = error
|

|	create frame for auto variables
|	and save this address

_BootCodeStart:

Entry:			movem.l	d2-d7/a2-a6,-(sp)
			move.l	4.w,a6
			lea	(-BASE_SIZE.w,sp),a5
			move.l	a5,sp

			lea	(BaseVar.w,pc),a0
			move.l	a5,(a0)

			move.l	a1,(BootRequest.w,a5)
			move.l	a6,(a5)			| SysBase

|	Set Up the Function Table

			lea	(FuncTable.w,a5),a0
			lea	(lilo_open.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_close.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_read.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_loadseg.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_unloadseg.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_exec.w,pc),a1
			move.l	a1,(a0)+

|	Allocate the File Buffer

			move.l	#TD_SECTOR,d0
			jbsr	SafeMalloc
			move.l	d0,(FileBuffer.w,a5)

|	Load and Execute the Boot Loader

BufferOK:		lea	(_LoaderVectorStart.w,pc),a0
			jbsr	lilo_loadseg
			move.l	d0,a2
			lea	(FuncTable.w,a5),a0
			move.l	a0,d0
			move.l	a2,a0
			jbsr	lilo_exec
			move.l	d0,d2
			move.l	a2,a0
			jbsr	lilo_unloadseg
			move.l	(FileBuffer.w,a5),a1
			jbsr	Free
			tst.l	d2
			jne	BootAmigaOS
BootError:		moveq	#-1,d0
			jbra	BootReturn

|	Boot the Standard Amiga Operating System

BootAmigaOS:		lea	(ExpansionName.w,pc),a1
			moveq	#37,d0
			jsr	(_LVOOpenLibrary,a6)
			tst.l	d0
			jeq	NoExpansion37
			move.l	d0,a1
			bset	#EBB_SILENTSTART,(eb_Flags,a1)
			jsr	(_LVOCloseLibrary,a6)
NoExpansion37:		lea	(DOSName.w,pc),a1
			jsr	(_LVOFindResident,a6)
			tst.l	d0
			jeq	BootError
			move.l	d0,a0
			move.l	(RT_INIT,a0),a0
			moveq	#0,d0
BootReturn:		lea	(BASE_SIZE.w,a5),sp
			movem.l	(sp)+,d2-d7/a2-a6
			rts


|
|	Open a File
|
|		in:	a0 = vector
|
|		out:	void
|

lilo_open:		move.l	(BaseVar.w,pc),a1
			move.l	a0,(FileVector.w,a1)
			clr.l	(FileOffset.w,a1)
			rts

|
|	Close a File
|
|		in:	void
|
|		out:	void
|

lilo_close:		move.l	a6,-(sp)
			move.l	(BaseVar.w,pc),a1
			move.l	(a1),a6			| SysBase
			move.l	(BootRequest.w,a1),a1
			move.w	#TD_MOTOR,(IO_COMMAND,a1)
			move.b	#IOF_QUICK,(IO_FLAGS,a1)
			clr.l	(IO_LENGTH,a1)
			jsr	(_LVODoIO,a6)
			move.l	(sp)+,a6
			rts

|
|	Read from the File
|
|		in:	a0 = dest
|			d0 = length
|
|		out:	void
|

lilo_read:		movem.l	d2-d3/a2/a5-a6,-(a7)
			move.l	(BaseVar.w,pc),a5
			move.l	(a5),a6			| SysBase
			move.l	d0,d2
			move.l	a0,a2
ReadLoop:		tst.l	d2
			jeq	End
			move.l	(FileOffset.w,a5),d3
			move.l	d3,d0
			and.l	#TD_SECTOR-1,d3
			jne	NoLoad
			moveq	#TD_SECSHIFT,d1
			asr.l	d1,d0
			move.l	(FileVector.w,a5),a0
			move.l	(4,a0,d0.l*4),d0
			jne	NoHole
			moveq	#TD_SECTOR/4-1,d1
			move.l	(FileBuffer.w,a5),a0
ClearLoop:		clr.l	(a0)+
			dbf	d1,ClearLoop
			jra	NoLoad
NoHole:			move.l	d0,a0
			move.l	(BootRequest.w,a5),a1
			move.w	#CMD_READ,(IO_COMMAND,a1)
			move.b	#IOF_QUICK,(IO_FLAGS,a1)
			move.l	#TD_SECTOR,(IO_LENGTH,a1)
			move.l	(FileBuffer.w,a5),(IO_DATA,a1)
			move.l	a0,(IO_OFFSET,a1)
			jsr	(_LVODoIO,a6)
			tst.l	d0
			jeq	NoLoad
			move.l	#READ_ERROR,d7
			jbra	Alert
NoLoad:			move.l	#TD_SECTOR,d0
			sub.l	d3,d0
			cmp.l	d2,d0
			jlt	NotAll
			move.l	d2,d0
NotAll:			move.l	(FileBuffer.w,a5),a0
			lea	(a0,d3.l),a0
			move.l	a2,a1
			move.l	d0,d3
			jsr	(_LVOCopyMem,a6)
			sub.l	d3,d2
			add.l	d3,a2
			add.l	d3,(FileOffset.w,a5)
			jra	ReadLoop
End:			movem.l	(a7)+,d2-d3/a2/a5-a6
			rts

|
|	Read a LONG from the File
|
|		in:	void
|
|		out:	d0 = data
|

ReadLong:		subq.l	#4,a7
			lea	(a7),a0
			moveq	#4,d0
			jbsr	lilo_read
			move.l	(a7)+,d0
back:			rts

|
|	memory allocation/deallocation
|	(made subroutines to conserve the tight space)
|

SafeMalloc:		move.l	#MEMF_PUB_CLR,d1
			jsr	(_LVOAllocVec,a6)
			tst.l	d0
			jne	back

			move.l	#NOT_ENOUGH_MEMORY,d7

Alert:			jmp	(_LVOAlert,a6)

Free:			jmp	(_LVOFreeVec,a6)

|
|	Load an AmigaOS Executable
|
|		in:	a0 = vector
|
|		out:	d0 = table
|

lilo_loadseg:		movem.l	d2-d5/a2-a3/a5-a6,-(a7)
			move.l	(BaseVar.w,pc),a5
			move.l	(a5),a6			| SysBase
			jbsr	lilo_open
			jbsr	ReadLong
			cmp.l	#HUNK_HEADER,d0
			jeq	HeaderOK

HunkError:		move.l	#HUNK_ERROR,d7
			jbra	Alert

HeaderOK:		jbsr	ReadLong
			tst.l	d0
			jne	HunkError
			jbsr	ReadLong
			move.l	d0,d5
			addq.l	#1,d0
			asl.l	#2,d0
			jbsr	SafeMalloc

TableOK:		move.l	d0,a2
			move.l	d5,(a2)
			jbsr	ReadLong
			tst.l	d0
			jne	HunkError
			jbsr	ReadLong
			addq.l	#1,d0
			cmp.l	d0,d5
			jne	HunkError
			moveq	#0,d2
AllocLoop:		jbsr	ReadLong
			asl.l	#2,d0
			jbsr	SafeMalloc
			move.l	d0,(4,a2,d2.l*4)
			addq.l	#1,d2
			cmp.l	d2,d5
			jne	AllocLoop
			moveq	#0,d2

NextHunk:		move.l	(4,a2,d2.l*4),a3
			jbsr	ReadLong
			sub.l	#HUNK_END,d0
			jeq	HunkEnd
			add.l	#HUNK_END-HUNK_DATA,d0
			jeq	HunkCodeData
			add.l	#HUNK_DATA-HUNK_CODE,d0
			jeq	HunkCodeData
			sub.l	#HUNK_BSS-HUNK_CODE,d0
			jeq	HunkBss
			sub.l	#HUNK_RELOC32-HUNK_BSS,d0
			jne	HunkError

HunkReloc32:		jbsr	ReadLong
			move.l	d0,d3
			jeq	NextHunk
			jbsr	ReadLong
			move.l	(4,a2,d0.l*4),d4
			subq.l	#1,d3
RelocLoop:		jbsr	ReadLong
			add.l	d4,(a3,d0.l)
			dbf	d3,RelocLoop
			jra	HunkReloc32
HunkCodeData:		jbsr	ReadLong
			move.l	a3,a0
			asl.l	#2,d0
			jbsr	lilo_read
			jra	NextHunk
HunkBss:		jbsr	ReadLong
			jra	NextHunk

HunkEnd:		addq.l	#1,d2
			cmp.l	d2,d5
			jne	NextHunk
			jbsr	lilo_close
			move.l	a2,d0
			movem.l	(a7)+,d2-d5/a2-a3/a5-a6
			rts

|
|	Unload an AmigaOS Executable
|
|		in:	a0 = table
|
|		out:	void
|

lilo_unloadseg:		movem.l	d2/a2-a3/a5-a6,-(a7)
			move.l	(BaseVar.w,pc),a5
			move.l	(a5),a6			| SysBase
			move.l	a0,a2
			move.l	a0,a3
			move.l	(a2)+,d2
FreeLoop:		move.l	(a2)+,a1
			jbsr	Free
			subq.l	#1,d2
			bhi.s	FreeLoop
			move.l	a3,a1
			jbsr	Free
			movem.l	(a7)+,d2/a2-a3/a5-a6
			rts


|
|	Execute an AmigaOS Executable
|
|		in:	a0 = table
|			d0 = args
|
|		out:	d0 = result
|

lilo_exec:		movem.l	d2/a2/a5-a6,-(a7)
			move.l	(BaseVar.w,pc),a5
			move.l	(a5),a6			| SysBase
			move.l	(4,a0),a2
			move.l	d0,d2
			lea	(_LVOCacheClearU.w,a6),a1
			cmp.w	#37,LIB_VERSION(a6)
			bcc.s	Cache37			| only after V37

			lea	(Cache36.w,pc),a1	| home-made cache flush

Cache37:		jsr	(a1)			| do the flush
			move.l	d2,d0			| pass args
			jsr	(a2)			| call loaded prg
			movem.l	(a7)+,d2/a2/a5-a6
			rts

Cache36:		movem.l	a5/a0-a1/d0-d1,-(sp)	| save all scratch (?)
			lea	(CacheSV.w,pc),a5
			jsr	(_LVOSuperVisor.w,a6)
			movem.l	(sp)+,a5/a0-a1/d0-d1

			rts

|
|	This code only runs on a pre2.0 (also early A3000 boot roms)
|	I assume there are still a lot of these (I have one too 8)
|
|	No need to check for '040, because this ROM doesn't work at all
|	on anything but a '030...
|

CacheSV:		.long	0x4e7a0002		| movec cacr,d0
			or.w	#0x808,d0		| clear '030 caches
			.long	0x4e7b0002		| movec.l d0,cacr
			rte

|
|	Some rearrangements must have been done..
|

|
|	Strings
|

ExpansionName:		.ascii	"expansion.library\0"
DOSName:		.ascii	"dos.library\0"


|
|	File Data for the Loader Program
|

			.align	4

_LoaderVectorStart:	.long	0			| File Size
			.long	0,0,0,0,0,0,0,0,0,0,0,0	| Data Blocks
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0,0,0
_LoaderVectorEnd:

BaseVar:		.ascii	"END!"

_BootCodeEnd:
