/*
 *		Amiga Linux/68k Loader
 *
 *		 Copyright 1995 by Geert Uytterhoeven
 *								 (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *		--------------------------------------------------------------------------
 *
 *		The original idea for this program comes from the Linux distribution:
 *
 *			LILO - Generic Boot Loader for Linux ("LInux LOader")
 *					 by Werner Almesberger
 *
 *		--------------------------------------------------------------------------
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of the Linux/68k
 *		distribution for more details.
 */


	/*
	 *		Function Table (passed at startup)
	 */

struct LiloTable {
	void *lilo_open;
	void *lilo_close;
	void *lilo_read;
	void *lilo_loadseg;
	void *lilo_unloadseg;
	void *lilo_exec;
};

extern struct LiloTable *LiloTable;


	/*
	 *		Open a File
	 */

static __inline void Open(u_long *vector)
{
	register void **a1 __asm("a1") = LiloTable->lilo_open;
	register u_long *a0 __asm("a0") = vector;

	__asm __volatile ("jsr (a1)"
							: /* no output */
							: "r" (a0), "r" (a1)
							: "a0","a1","d0","d1","memory");
}


	/*
	 *		Close the File
	 */

static __inline void Close(void)
{
	register void **a1 __asm("a1") = LiloTable->lilo_close;

	__asm __volatile ("jsr (a1)"
							: /* no output */
							: "r" (a1)
							: "a0","a1","d0","d1","memory");
}


	/*
	 *		Read a Block from the File
	 */

static __inline void Read(void *dest, u_long length)
{
	register void **a1 __asm("a1") = LiloTable->lilo_read;
	register void *a0 __asm("a0") = dest;
	register u_long d0 __asm("d0") = length;

	__asm __volatile ("jsr (a1)"
							: /* no output */
							: "r" (a0), "r" (a1), "r" (d0)
							: "a0","a1","d0","d1","memory");
}


	/*
	 *		Load an AmigaOS Executable
	 */

static __inline u_long *LoadSeg(u_long *vector)
{
	register u_long *_res __asm("d0");
	register void **a1 __asm("a1") = LiloTable->lilo_loadseg;
	register u_long *a0 __asm("a0") = vector;

	__asm __volatile ("jsr (a1)"
							: "=r" (_res)
							: "r" (a0), "r" (a1)
							: "a0","a1","d0","d1","memory");
	return(_res);
}


	/*
	 *		Unload an AmigaOS Executable
	 */

static __inline void UnLoadSeg(u_long *table)
{
	register void **a1 __asm("a1") = LiloTable->lilo_unloadseg;
	register u_long *a0 __asm("a0") = table;

	__asm __volatile ("jsr (a1)"
							: /* no output */
							: "r" (a0), "r" (a1)
							: "a0","a1","d0","d1","memory");
}


	/*
	 *		Execute an AmigaOS Executable
	 */

static __inline long Exec(u_long *table, long arg)
{
	register long _res __asm("d0");
	register void **a1 __asm("a1") = LiloTable->lilo_exec;
	register u_long *a0 __asm("a0") = table;
	register long d0 __asm("d0") = arg;

	__asm __volatile ("jsr (a1)"
							: "=r" (_res)
							: "r" (a0), "r" (a1), "r" (d0)
							: "a0","a1","d0","d1","memory");
	return(_res);
}
